/***********************************************************
        Copyright 1996 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <net/route.h>
#include <netdb.h>
#include <syslog.h>
#include <errno.h>


#include "defs.h"
#include "bootp.h"
#include "bootpd.h"
#include "dhcp.h"
#include "dhcpd.h"
#include "dynamic.h"
#include "readfile.h"
#include "perm.h"


extern short bootpc_port;
extern short bootps_port;
extern int debug,s,pktlen;
hash_tbl *boundhshtbl;
hash_tbl *ipboundhshtbl;


struct unknownhost *unknown;
unsigned int num_unknown;

extern struct in_addr my_ip_addr;
static int i;
int init=0;
unsigned long dhcp_alarm;
unsigned int leases;
char buf[256];
static unsigned char send_buf[2000];
static struct bootp *rpkt;

extern struct sockaddr_in recv_addr;
void send_dhcp_reply();
void send_dhcpnack();
void order_insert();
struct host *create_stub __P((struct host* hp, struct client_info *cl));
struct host *create_lease __P((struct host* hp, struct shared_bindata *cliid,
				struct shared_bindata *vend, 
				struct shared_bindata *user, unsigned secs));
struct host *hp;
struct shared_bindata *cliid,*vend,*user;
unsigned secs;
struct flag do_req_opts();
unsigned int discovers_received,offers_sent;
unsigned int requests_received,acks_sent,nacks_sent;
unsigned int ips_expired,ips_reclaimed,declines_received;
unsigned int releases_received;
unsigned int informs_received;
unsigned long gcinterval;
unsigned int bootp_requests_received,bootp_replys_sent;

unsigned long now,starttime;
struct in_addr my_ip_addr;
int pktlen;
void report_cliid();

char *cctime(at)
unsigned long at;
{
 char *r;

 r=ctime((time_t*)&at);
 r[strlen(r)-1]=0;
 return r;
}

void init_dhcp()
{
 struct rtentry rt;
 struct sockaddr_in addr;
 struct lastbind *binding;
 struct client_info cl;
 hash_datum *data;
 int len,i,rc;
 unsigned hashcode;
 struct host *hp;
 struct in_addr tmpaddr;
 char flag;

 lastbindhshtbl=hash_Init(HASHTABLESIZE);
 boundhshtbl=hash_Init(HASHTABLESIZE);
 ipboundhshtbl=hash_Init(HASHTABLESIZE);
 cache2disk=hash_Init(HASHTABLESIZE);

 if(debug)
	report(LOG_INFO,"using last binds dir %s",bindingsdir);
 if((lastbindpermanent=hashf_Init(bindingsdir,HASHTABLESIZE)) == NULL){
	report(LOG_ERR,"Could not open last binds dir : %s",LASTBINDDIR);
	exit(1);
 }

 hash_Reset(boundhshtbl,free_host);
 hash_Reset(ipboundhshtbl,free_host);
 hash_Reset(lastbindhshtbl,free_lastbind);
 hash_Reset(cache2disk,free_lastbind);

 i=0;
 leases=0;
 unknown=NULL;
 now=time(0);
 tmpaddr.s_addr=0xffffffff;
 for(data=hashf_FirstEntry(lastbindpermanent,&len);data != NULL;
			data=hashf_NextEntry(lastbindpermanent,&len)){
	binding=data2bind(data,len);
	if(!binding){
		report(LOG_ERR,"Corrupt entry found on disk");
		continue;
		}
	binding->linkcount++;
	if(debug>4){
		report_cliid(LOG_INFO,binding->cliid);
		report(LOG_INFO,"ip=%s replys=%d  at %s",inet_ntoa(binding->last_ip),
			binding->replys,cctime(binding->last_touch));
	}
	hash_Insert(lastbindhshtbl,
		hash_HashFunction(binding->cliid->data,binding->cliid->length),
		lastbindcmp,binding,binding);

	i++;
	if(now<binding->expire && binding->code!=SYNCH_FLAG &&
		binding->code!=ROAMING_FLAG && binding->expire!=INFINITY){ 
		/* recover binding */
		hashcode=hash_HashFunction((u_char*)&binding->last_ip,4);
		hp=(struct host *)hash_Lookup(iphashtable,hashcode,ipcmp,
					&binding->last_ip);
		if(!hp)
			continue;
		if(binding->code==BOOTP_FLAG && !hp->flags.dynamic)
			continue;

		bzero(&cl,sizeof(cl));
		cl.cliid=binding->cliid;
		cl.vend=binding->vend;
		cl.user=binding->user;
		cl.clhashcode=hash_HashFunction(binding->cliid->data,binding->cliid->length);
		hp=create_stub(hp,&cl);
		if(!create_lease(hp,cl.cliid,cl.vend,cl.user,0)){
			hp->linkcount=1;
			free_host(hp);
			continue;
		}
		hp->expire=binding->expire;
		hp->code=binding->code;
		if(hp->code==RESERVED_FLAG)
			flag=ANY;
		if(hp->code==DHCP_FLAG)
			flag=DHCP_ACCESS;
		if(hp->code==BOOTP_FLAG)
			flag=BOOTP_ACCESS;
		
		if(hash_Exists(clhashtable, cl.clhashcode,cliidcmp,cl.cliid))
			flag|=REG_ACCESS;
		hp->linkcount=1;
		if(hp->flags.dynamic && !use_addr(hp,tmpaddr,flag,1)){
			free_host(hp);
			continue;
		}
		if(hash_Insert(ipboundhshtbl,hp->iphashcode,ipcmp,&hp->iaddr,hp)){
			report(LOG_ERR,"duplicate entries for %s, database corrupted",inet_ntoa(hp->iaddr));
			free_host(hp);
			continue;
		}
		hp->linkcount++;
		hash_Insert(boundhshtbl,cl.clhashcode,boundcliidcmp,hp->cl,hp);
		leases++;
		if(debug>2)
			report(LOG_INFO,"Recoverd %s %s",inet_ntoa(hp->iaddr),cctime(binding->expire));
	}

		
 }
 if(lastbindpermanent->next){
	fclose(lastbindpermanent->next);
	lastbindpermanent->next=NULL;
 }
 if(debug>2)
 	report(LOG_INFO,"%d bindings retrieved from disk",i);
 if(max_cachequeue==0)
 	max_cachequeue=20;
 if(flush_interval==0)
 	flush_interval=30*60L;
 if(gcinterval==0)
	gcinterval=30*60L;
 dhcp_alarm=time(0)+gcinterval;
 flush_alarm=time(0)+flush_interval;
 num_unknown=0;
 informs_received=0;
 discovers_received=0;
 requests_received=0;
 releases_received=0;
 declines_received=0;
 ips_reclaimed=0;
 ips_expired=0;
 acks_sent=0;
 nacks_sent=0;
 offers_sent=0;

}

/* if this is a dhcp packet, return the type otherwise 0
   if the options in this packet are poorly formed, return -1 
 */
char dhcp_msg_type(pkt)
struct bootp *pkt;
{
 u_char *cp;
 unsigned hashcode;


 if(bcmp(pkt->bp_vend,vm_rfc1048,4))
	return 0;
 for(cp=(u_char*)pkt,i=4+((u_long)pkt->bp_vend-(u_long)pkt); i<pktlen ;){
	if(cp[i]==TAG_PAD){
		i++;
		continue;
	}
	if(cp[i]==TAG_END)
			return 0;

	if(cp[i]==TAG_DHCP_MSG){
		i+=2;
		if(i>=pktlen || cp[i]>DHCPINFORM || cp[i]<DHCPDISCOVER)
			return -1;
		return cp[i];
	}
		
	i++;
	if(i==pktlen || ((int)(cp[i]+i)>=pktlen))
		return -1;
	i+=cp[i]+1;
  }
  return 0;
}

void report_cliid(log,cl)
int log;
struct shared_bindata *cl;
{
	int rc;
	for(i=0,rc=1;i<cl->length && rc;i++)
		if(!(ispunct(cl->data[i])||isalnum(cl->data[i])||isspace(cl->data[i])))
		 rc=0;
	if(rc){
		bcopy(cl->data,buf,cl->length);
		buf[cl->length]=0;
	} else
		for(i=0,buf[0]=0;i<cl->length;i++)
			sprintf(buf+strlen(buf),"%02X ",cl->data[i]);
		
	report(log,"Client id: %s len=%d",buf,cl->length);
}

int dhcp_opts(pkt,cl)
struct bootp *pkt;
struct client_info *cl;
{
	u_char *cp,overload;
	int j;

	overload=0;
	cl->cliidhw=(struct shared_bindata*)
			malloc(sizeof(struct shared_bindata)+pkt->bp_hlen);
	cl->cliidhw->length=haddrlength(pkt->bp_htype)+1;
	if (cl->cliidhw->length-1 != pkt->bp_hlen) {
		report(LOG_NOTICE, "bad addr len from from %s address %s",
		netname(pkt->bp_htype),
		haddrtoa(pkt->bp_chaddr,cl->cliidhw->length-1 ));
	}
	cl->cliidhw->data[0]=pkt->bp_htype;
	cl->cliidhw->linkcount=1;
	bcopy(pkt->bp_chaddr,cl->cliidhw->data+1,cl->cliidhw->length-1); 

	for(i=VEND_OFFSET,cp=(u_char*)pkt;i<pktlen;){
		switch(cp[i]){
		case TAG_PAD:i++;continue;
		case TAG_DHCP_OVL:
			overload=cp[i+2]&3;
		case TAG_END:
			if(overload&1){  /* file overload */
				i=FILE_OFFSET;
				continue;
			}
			if(overload&2){  /* sname overload */
				i=SNAME_OFFSET;
				continue;
			}
			if(!cl->cliid){
				cl->cliid=cl->cliidhw;
				cl->cliid->linkcount++;
			}
			if(!cl->req_addr.s_addr)
				cl->req_addr.s_addr=pkt->bp_ciaddr.s_addr;
			return 0;
		case TAG_MESSAGE:
			cl->msg=(char*)malloc(cp[i+1]+1);
			bcopy(cp+i+2,cl->msg,cp[i+1]);
			cl->msg[cp[i+1]+1]=0;
			break;
		case TAG_DHCP_USERCLASS:
			cl->user=(struct shared_bindata*)
					malloc(sizeof(struct shared_bindata)+cp[i+1]-1);
			bcopy(&cp[i+2],cl->user->data,cp[i+1]);
			cl->user->linkcount=1;
			cl->user->length=cp[i+1];
			break;
		case TAG_VENDOR:
			cl->vend=(struct shared_bindata*)
					malloc(sizeof(struct shared_bindata)+cp[i+1]-1);
			bcopy(&cp[i+2],cl->vend->data,cp[i+1]);
			cl->vend->linkcount=1;
			cl->vend->length=cp[i+1];
			break;
		case TAG_DHCP_SERVERID: bcopy(&cp[i+2],&cl->serv_addr.s_addr,4); break;
		case TAG_DHCP_IP: bcopy(&cp[i+2],&cl->req_addr.s_addr,4); break;
		case TAG_MAX_MSGSZ: bcopy(&cp[i+2],&cl->max_len,4); break;
		case TAG_DHCP_CLIID:
			cl->cliid=(struct shared_bindata*)
					malloc(sizeof(struct shared_bindata)+cp[i+1]-1);
			bzero(cl->cliid,sizeof(struct shared_bindata)+cp[i+1]-1);
			bcopy(&cp[i+2],cl->cliid->data,cp[i+1]);
			cl->cliid->length=cp[i+1];
			cl->cliid->linkcount=1;
			break;
		case TAG_DHCP_REQ:
			cl->req_opts=(unsigned char*)malloc(cp[i+1]+1);
			bcopy(&cp[i+1],cl->req_opts,cp[i+1]+1);
			break;
		default:break;
		}
		i++;
		i+=cp[i]+1;
	}
	report(LOG_NOTICE,"an Error occured while parsing options");
	return 1; 
}

boolean ipcmp(a, b)
hash_datum *a,*b;
{
char *key;
struct host *data;
	key=(char *)a;
	data=(struct host *)b;
	return (((struct in_addr*)key)->s_addr==data->iaddr.s_addr);
}
boolean cliidhwlookcmp(a, b)
hash_datum *a,*b;
{
struct shared_bindata *key;
struct host *data;
	key=(struct shared_bindata *)a;
	data=(struct host *)b;
	if(data->htype!=key->data[0])
		return FALSE;

	if(bcmp(data->haddr,&key->data[1],key->length-1))
		return FALSE;
	return TRUE;
}
boolean cliidcmp(a,b)
hash_datum *a,*b;
{
struct shared_bindata *key;
struct host *data;
	key=(struct shared_bindata *)a;
	data=(struct host *)b;
	if(data->flags.cliid)
		return (key->length==data->cliid->length && 
			!bcmp(key->data,data->cliid->data,key->length));
	else
		return FALSE;
}
boolean boundcliidcmp(a,b)
hash_datum *a,*b;
{
struct shared_bindata *key;
struct host *data;
	key=(struct shared_bindata *)a;
	data=(struct host *)b;
	return (key->length==data->cl->length && 
			!bcmp(key->data,data->cl->data,key->length));
}
	
void free_client_info(cl)
struct client_info *cl;
{
#if 0
	if(--cl->linkcount)
		return;
#endif
	if(cl->req_opts)free(cl->req_opts);
	del_bindata(cl->cliid);
	del_bindata(cl->cliidhw);
	del_bindata(cl->vend);
	del_bindata(cl->user);
#if 0
	free(cl);
#endif
}

void dhcp_discover(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp, *registered;
	int32 dest;
	unsigned len,iphashcode,hwhashcode,rc;
	struct opts opt;
	struct lastbind *oldbind;
	int32 bootsize=0;
	int exists;

	bzero(&cl,sizeof(cl));
	cl.max_len=576;

	if(debug>1)
		report(LOG_INFO,"Received DHCPDISCOVER message");
	discovers_received++;
	rc=dhcp_opts(pkt,&cl);
	if(cl.msg)free(cl.msg);

	if(!cl.cliid)
		report(LOG_ERR,"Couldn't determine client identifier");
	if(rc == -1){
		report(LOG_ERR,"Error in parsing options");
		free_client_info(&cl);
		return;
	}

	if(debug>2)
		report_cliid(LOG_INFO,cl.cliid);


	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	registered=hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
	hp =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);
	if(hp)
		exists=TRUE;
	else
		exists=FALSE;
	if(!hp){
		if(debug>2)
			report(LOG_INFO,"Seeing if oldbinding exists and available");
		oldbind=(struct lastbind*)hash_Lookup(lastbindhshtbl,cl.clhashcode,lastbindcmp,&cl);
		if(oldbind){
			if(debug>2)
				report(LOG_INFO,"Checking ip %s",inet_ntoa(oldbind->last_ip));
			iphashcode=hash_HashFunction((u_char*)&oldbind->last_ip.s_addr,4);
			hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp,
					(hash_datum*)&oldbind->last_ip.s_addr);

			if(hp && hp->flags.cliid && !BINDATA_EQ(hp->cliid,cl.cliid))
						hp=NULL;
			if(hp && hp->flags.dynamic)
				if(registered)
					hp=use_addr(hp,pkt->bp_giaddr,REG_ACCESS|DHCP_ACCESS,0);
				else
					hp=use_addr(hp,pkt->bp_giaddr,DHCP_ACCESS,0);
		}
		if(hp && !check_net(pkt->bp_giaddr,hp->iaddr))
			hp=NULL;
	}
	if(!hp){
		if(debug>2)
			report(LOG_INFO, "See is requested ip availible %s",
					inet_ntoa(cl.req_addr));
		if(cl.req_addr.s_addr!=0){
			iphashcode=hash_HashFunction((u_char*)&cl.req_addr.s_addr,4);
			hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp,
					(hash_datum*)&cl.req_addr.s_addr);
			if(hp &&  hp->flags.cliid && !BINDATA_EQ(hp->cliid,cl.cliid))
						hp=NULL;
			if(hp && hp->flags.dynamic)
				if(registered)
					hp=use_addr(hp,pkt->bp_giaddr,REG_ACCESS|DHCP_ACCESS,0);
				else
					hp=use_addr(hp,pkt->bp_giaddr,DHCP_ACCESS,0);
			
		}
		if(hp && !check_net(pkt->bp_giaddr,hp->iaddr))
			hp=NULL;
	}

	if(!hp){
		report(LOG_INFO,"IP not availible, looking client up in database");

		hp =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
		if(!hp){
			hwhashcode=hash_HashFunction(cl.cliidhw->data+1,cl.cliidhw->length-1);
			hp =(struct host *)hash_Lookup(hwhashtable, hwhashcode, 
				cliidhwlookcmp,cl.cliidhw);
		}
		if(hp && !check_net(pkt->bp_giaddr,hp->iaddr))
				hp=NULL;
	}

	if(!hp){
		if(registered)
			hp=get_dynamic_ip(pkt->bp_giaddr,REG_ACCESS|DHCP_ACCESS);
		else
			hp=get_dynamic_ip(pkt->bp_giaddr,DHCP_ACCESS);
	}

	if(!hp && registered && registered->flags.roaming)
			hp=get_dynamic_ip(pkt->bp_giaddr,DHCP_ACCESS|ROAMING_ACCESS);

	if(!hp){
		report(LOG_INFO,"no dynamic ip's availible, ignoring");
		free_client_info(&cl);
		return;
	}

	report(LOG_INFO,"Found client %s (%s)",inet_ntoa(hp->iaddr),
					hp->hostname->string);
	if(!exists){
		hp=create_stub(hp,&cl);
		if(hp->flags.dynamic)
			hp->code=DHCP_FLAG;
		else
			hp->code=RESERVED_FLAG;
		hp->expire=now+TIME_OFFSET;
		hp->xid=pkt->bp_xid;
		hp->linkcount+=2;
		if(!hash_Insert(boundhshtbl,cl.clhashcode, boundcliidcmp, cl.cliid, (hash_datum*)hp))
		leases++;
		hash_Insert(ipboundhshtbl,hp->iphashcode,ipcmp,&hp->iaddr,(hash_datum*)hp);
	}
	rpkt=(struct bootp *)send_buf;
	bcopy(pkt,rpkt,pktlen);
	rpkt->bp_op=BOOTREPLY;
	rpkt->bp_hops=0;
	rpkt->bp_secs=0;
	rpkt->bp_yiaddr.s_addr=hp->iaddr.s_addr;
    	rpkt->bp_ciaddr.s_addr=0;
	if(fill_header(rpkt,hp,&bootsize)){
		free_client_info(&cl);
		reuse_addr(hp);
		return;
	}


	dest=hp->flags.reply_addr?hp->reply_addr.s_addr:0;
	if((rpkt->bp_flags&0x8000) && !rpkt->bp_giaddr.s_addr)
		dest=0xffffffff;
	if(dest==0xffffffff)
		rpkt->bp_flags|=0x8000;

	opt.cur=0;
	opt.len=cl.max_len;
	insert_byte_opt(rpkt,&opt,TAG_DHCP_MSG,DHCPOFFER);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_SERVERID,my_ip_addr.s_addr);
	if(hp->flags.lease)
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_IPLEASE,hp->lease);
	if(hp->flags.renew)
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_RENEW,hp->renew);
	if(hp->flags.rebind)
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_REBIND,hp->rebind);
	if(cl.user){
		insert_string_opt(rpkt,&opt,TAG_DHCP_USERCLASS,
				cl.user->data,cl.user->length);
	}
	insert_end_opt(rpkt,&opt);

	report(LOG_INFO,"Sending DHCPOFFER (%s) message max_len=%d left=%d"
			,inet_ntoa(hp->iaddr),opt.len,opt.len-opt.cur);
	offers_sent++;
	len=300;
	if(opt.cur<VEND_OFFSET)
		len=cl.max_len;
	if(opt.cur>len)
		len=opt.cur;
	send_dhcp_reply(rpkt,len,dest);
	del_bindata(cl.cliid);
	del_bindata(cl.cliidhw);
	del_bindata(cl.vend);
	del_bindata(cl.user);
}

void dhcp_request(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2,*t;
	unsigned int len,dest;
	struct opts opt;
	struct flag flags;
	unsigned iphashcode,hwhashcode,rc;
	int32 bootsize=0;
	int registered;

	bzero(&cl,sizeof(cl));
	cl.max_len=576;
	hp2=NULL;

	if(debug>1)
		report(LOG_INFO,"Received DHCPREQUEST message");
	requests_received++;
	rc=dhcp_opts(pkt,&cl);
	if(cl.msg)free(cl.msg);

	if(!cl.cliid)
		report(LOG_ERR,"Couldn't determine client identifier");
	if(rc == -1){
		report(LOG_ERR,"Error in parsing options");
		free_client_info(&cl);
		return;
	}

	if(debug>2)
		report_cliid(LOG_INFO,cl.cliid);

	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	hp =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);
	registered=hash_Exists(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);

	if(hp){
		if(debug>2)
			report(LOG_INFO,"Found lease and extending if necessary");
	} else if(cl.req_addr.s_addr!=0){
		if(debug>2)
			report(LOG_INFO,"Client is not renewing");
		iphashcode=hash_HashFunction((u_char*)&cl.req_addr.s_addr,4);
		hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp,
					(hash_datum*)&cl.req_addr.s_addr);
		if(hp && !hp->flags.dynamic){
		if(hp && hp->flags.cliid && !CLIID_EQ(hp->cliid,cl.cliid)){
			if(debug>2)
				report(LOG_INFO,
					"Client Identifier doesn't match (%s), access denied",
						inet_ntoa(cl.req_addr));
			send_dhcpnack(pkt,cl.cliid,"Access denied",cl.max_len);
			free_client_info(&cl);
			return;
		} else
		if(hp && hp->flags.haddr && hp->flags.htype)
			if((hp->htype!=cl.cliid->data[0]) || ((cl.cliid->length-1)!=haddrlength(hp->htype))
					|| bcmp(hp->haddr,cl.cliid->data+1,cl.cliid->length-1)){ 
				if(debug>2)
					report(LOG_INFO,
						"Hardware address doesn't match, access denied");
				send_dhcpnack(pkt,cl.cliid,"Access denied",cl.max_len);
				free_client_info(&cl);
				return;
			}
		}
		if(hp && !check_net(pkt->bp_giaddr,hp->iaddr))
				hp=NULL;
		else if(registered)
				hp=use_addr(hp,pkt->bp_giaddr,REG_ACCESS|DHCP_ACCESS,0);
			else
				hp=use_addr(hp,pkt->bp_giaddr,DHCP_ACCESS,0);
		
		if(hp){
			hp=create_stub(hp,&cl);
			hp->linkcount+=2;
			hp->xid=pkt->bp_xid;
			leases++;
			hash_Insert(boundhshtbl,cl.clhashcode, boundcliidcmp, cl.cliid,
						(hash_datum*)hp);
			hash_Insert(ipboundhshtbl,hp->iphashcode, ipcmp, &hp->iaddr.s_addr,
				(hash_datum*)hp);
		}
	}
	if(!hp){
		if(debug>2)
			report(LOG_INFO,"Not reqesting extension ");
		hp =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
		if(!hp){
			hwhashcode=hash_HashFunction(cl.cliid->data+1,cl.cliid->length-1);
			hp =(struct host *)hash_Lookup(hwhashtable, hwhashcode, 
				cliidhwlookcmp,cl.cliid);
		}
		if(hp && !check_net(pkt->bp_giaddr,hp->iaddr))
				hp=NULL;
		if(hp){
			hp=create_stub(hp,&cl);
			hp->linkcount+=2;
			hp->xid=pkt->bp_xid;
			leases++;
			hash_Insert(boundhshtbl,cl.clhashcode, boundcliidcmp, cl.cliid,
				(hash_datum*)hp);
			hash_Insert(ipboundhshtbl,hp->iphashcode, ipcmp, &hp->iaddr.s_addr, 
				(hash_datum*)hp);
		}
	}
	if(!hp){
		report(LOG_INFO,"Unknown client, or access denied");
		if(cl.serv_addr.s_addr==my_ip_addr.s_addr || cl.serv_addr.s_addr==0)
			send_dhcpnack(pkt,cl.cliid,"Database has changed",cl.max_len);
		free_client_info(&cl);
		return;
	}

	if(!hp->flags.complete)
		hp=create_lease(hp,cl.cliid,cl.vend,cl.user,cl.secs);
	if(hp->flags.dynamic)
		hp->code=DHCP_FLAG;
	else
		hp->code=RESERVED_FLAG;
 
	if(hp->lease!=INFINITY){
		hp->expire=now+hp->lease;
	}
	report(LOG_INFO,"Found client %s (%s)",inet_ntoa(hp->iaddr),
					hp->hostname->string);
	if(cl.serv_addr.s_addr==my_ip_addr.s_addr || cl.serv_addr.s_addr==0){
    rpkt=(struct bootp *)send_buf;
    bcopy(pkt,rpkt,pktlen);
    rpkt->bp_op=BOOTREPLY;
    rpkt->bp_hops=0;
    rpkt->bp_secs=0;
    rpkt->bp_yiaddr.s_addr=hp->iaddr.s_addr;
    rpkt->bp_siaddr.s_addr = hp->bootserver.s_addr;
	if(fill_header(rpkt,hp,&bootsize)){
		free_client_info(&cl);
		reuse_addr(hp);
		return;
	}
	dest=hp->flags.reply_addr?hp->reply_addr.s_addr:0;
	if((rpkt->bp_flags&0x8000) && !rpkt->bp_giaddr.s_addr)
		dest=0xffffffff;
    if(dest==0xffffffff)
		rpkt->bp_flags|=0x8000;

	opt.cur=0;
	opt.len=cl.max_len;
	insert_byte_opt(rpkt,&opt,TAG_DHCP_MSG,DHCPACK);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_SERVERID,my_ip_addr.s_addr);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_IPLEASE,hp->lease);
	if(hp->flags.renew)
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_RENEW,hp->renew);
	if(hp->flags.rebind)
		insert_ulong_opt(rpkt,&opt,TAG_DHCP_REBIND,hp->rebind);
	if(cl.user){
		insert_string_opt(rpkt,&opt,TAG_DHCP_USERCLASS,
					cl.user->data,cl.user->length);
	}
    flags=do_req_opts(hp,cl.req_opts,rpkt,&opt);
	if(hp->flags.not_rq)
		do_opts(hp,flags,rpkt,&opt);
	insert_end_opt(rpkt,&opt);
		report(LOG_INFO,"Sending DHCPACK message max_len=%d left=%d"
	    	,opt.len,opt.len-opt.cur);
		acks_sent++;
	len=300;
	if(opt.cur<VEND_OFFSET)
		len=cl.max_len;
	if(opt.cur>len)
		len=opt.cur;
		send_dhcp_reply(rpkt,len,dest);
	}else if(debug>2)
		report(LOG_INFO,"Intended for %s, not ACKING",inet_ntoa(cl.serv_addr));
	update_lastbind(&cl,hp);
	free_client_info(&cl);

}

void dhcp_release(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2;
	int32 dest;
	unsigned iphashcode,hwhashcode,rc;

	bzero(&cl,sizeof(cl));
	cl.max_len=576;

	if(debug>1)
		report(LOG_INFO,"Received DHCPRELEASE message");
	releases_received++;
	rc=dhcp_opts(pkt,&cl);

	if(cl.serv_addr.s_addr!=0 && cl.serv_addr.s_addr != my_ip_addr.s_addr){
		if(debug>2)
			report(LOG_INFO,"Message is for server %s",inet_ntoa(my_ip_addr));
		free(cl.cliid);
		if(cl.req_opts)free(cl.req_opts);
		return;
	}

	if(!cl.cliid)
		report(LOG_ERR,"Couldn't determine client identifier");
	if(rc == -1){
		report(LOG_ERR,"Error in parsing options");
		if(cl.req_opts)free(cl.req_opts);
		del_bindata(cl.cliid);
		del_bindata(cl.cliidhw);
		del_bindata(cl.vend);
		del_bindata(cl.user);
		return;
	}

	if(debug>2)
		report_cliid(LOG_INFO,cl.cliid);

	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	hp =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);
	if(!hp){
		if(debug>1)
			report(LOG_ERR,"Binding doesn't exist for this client");
		if(cl.req_opts)free(cl.req_opts);
		del_bindata(cl.cliid);
		del_bindata(cl.cliidhw);
		del_bindata(cl.vend);
		del_bindata(cl.user);
		return;
	}

	if(debug>2)
		report(LOG_INFO,"Freeing binding");
	if(cl.msg)
		report(LOG_INFO,"Message from client %s",cl.msg);
	reuse_addr(hp);

	if(cl.req_opts)free(cl.req_opts);
	del_bindata(cl.cliid);
	del_bindata(cl.cliidhw);
	del_bindata(cl.vend);
	del_bindata(cl.user);

	return;
}

void dhcp_decline(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2;
	int32 dest;
	struct order_list **p;
	unsigned rc;

	bzero(&cl,sizeof(cl));
	cl.max_len=576;
	if(debug>1)
	rc=dhcp_opts(pkt,&cl);
	if(debug>1)
		report(LOG_INFO,"Received DHCPDECLINE message");
	declines_received++;
	if(debug>2)
		report_cliid(LOG_INFO,cl.cliid);
	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	hp =(struct host *)hash_Lookup(boundhshtbl,cl.clhashcode,boundcliidcmp,cl.cliid);

	if(hp)
		reuse_addr(hp);
	if(cl.req_opts)free(cl.req_opts);
	del_bindata(cl.cliid);
	del_bindata(cl.cliidhw);
	del_bindata(cl.vend);
	del_bindata(cl.user);

}

void dhcp_inform(pkt)
struct bootp *pkt;
{
	struct client_info cl;
	struct host *hp,*hp2,generic;
	struct opts opt;
	char override[2000];
	int len;
	unsigned rc,iphashcode,hwhashcode,nmhashcode;


	bzero(&cl,sizeof(cl));
	cl.max_len=576;
	rc=dhcp_opts(pkt,&cl);
	if(debug>1)
		report(LOG_INFO,"Received DHCPINFORM message");
	informs_received++;
	if(debug>2)
		report_cliid(LOG_INFO,cl.cliid);


	cl.clhashcode=hash_HashFunction(cl.cliid->data,cl.cliid->length);
	hp =(struct host *)hash_Lookup(clhashtable, cl.clhashcode,cliidcmp,cl.cliid);
	if(!hp){
		hwhashcode=hash_HashFunction(cl.cliid->data+1,cl.cliid->length-1);
		hp =(struct host *)hash_Lookup(hwhashtable, hwhashcode, 
			cliidhwlookcmp,cl.cliid);
	}
	if(!hp){
		iphashcode=hash_HashFunction((u_char*)&pkt->bp_ciaddr.s_addr,4);
		hp=(struct host *)hash_Lookup(iphashtable,iphashcode,ipcmp,
						&pkt->bp_ciaddr.s_addr);
	}

	if(!hp){
		if(debug>1)
			report(LOG_INFO,"Client is not registered, using generic response");
		if(cl.user){
			override[0]='$';
			override[1]='C';
			override[2]=0;
			for(i=0;i<cl.user->length;i++)
				sprintf(override+strlen(override),"%02X",cl.user->data[i]);
		} else if(cl.vend){
			override[0]='$';
			override[1]='V';
			override[2]=0;
			for(i=0;i<cl.user->length;i++)
				sprintf(override+strlen(override),"%02X",cl.vend->data[i]);
		} else 
			strcpy(override,"$DHCP");
		bzero(&generic,sizeof(generic));
		nmhashcode=hash_HashFunction((u_char*)override,strlen(override));
		hp2=(struct host *)hash_Lookup(nmhashtable,nmhashcode,nmcmp,override);
		if(hp2)
			fill_defaults2(&generic,hp2);
		else{
			if(cl.req_opts)free(cl.req_opts);
			del_bindata(cl.cliid);
			del_bindata(cl.cliidhw);
			del_bindata(cl.vend);
			del_bindata(cl.user);
			return;
		}

	}
		
	hp=create_lease(hp,cl.cliid,cl.vend,cl.user,0);
	rpkt=(struct bootp *)malloc(cl.max_len);
	bcopy(pkt,rpkt,pktlen);
	rpkt->bp_op=BOOTREPLY;
	rpkt->bp_hops=0;
	rpkt->bp_secs=0;
	rpkt->bp_yiaddr.s_addr=rpkt->bp_ciaddr.s_addr;
	rpkt->bp_siaddr.s_addr = hp->bootserver.s_addr;

	opt.cur=0;
	opt.len=cl.max_len;
	insert_byte_opt(rpkt,&opt,TAG_DHCP_MSG,DHCPACK);
	insert_ulong_opt(rpkt,&opt,TAG_DHCP_SERVERID,my_ip_addr.s_addr);
	if(cl.user){
		insert_string_opt(rpkt,&opt,TAG_DHCP_USERCLASS,
					cl.user->data,cl.user->length);
		del_bindata(cl.user);
	}
	if(cl.vend)
		del_bindata(cl.vend);
	insert_end_opt(rpkt,&opt);

	report(LOG_INFO,"Sending DHCPACK message max_len=%d left=%d" 
		,opt.len,opt.len-opt.cur);
	acks_sent++;
	len=300;
	if(opt.cur<VEND_OFFSET)
		len=cl.max_len;
	if(opt.cur>len)
		len=opt.cur;
	send_dhcp_reply(rpkt,len,rpkt->bp_ciaddr.s_addr);

	if(cl.req_opts)free(cl.req_opts);
	del_bindata(cl.cliid);
	del_bindata(cl.cliidhw);
	del_bindata(cl.vend);
	del_bindata(cl.user);
	if(&generic!=hp)
		free(hp);
}



int do_opts(hp,flags,bp,cur)
struct host *hp;
struct flag flags;
u_char *bp;
struct opts *cur;
{
 return 0;
}

struct flag do_req_opts(hp,req_opts,bp,cur)
struct host *hp;
u_char *req_opts,*bp;
struct opts *cur;
{
	int i,j,rc;
	struct flag f;

	bzero(&f,sizeof(f));
	if(!req_opts)
		return f;
	for(i=1;i<(int)(req_opts[0]+1);i++){
		switch(req_opts[i]){                       
		case TAG_NODE_TYPE:
			if(f.node_type=hp->flags.node_type)
			rc=insert_byte_opt(bp,cur,TAG_NODE_TYPE,hp->node_type);
			break;
		case TAG_DHCP_REBIND:
			if(f.rebind=hp->flags.rebind)
			rc=insert_ulong_opt(bp,cur,TAG_DHCP_REBIND,hp->rebind);
			break;
		case TAG_DHCP_RENEW:
			if(f.renew=hp->flags.renew)
			rc=insert_ulong_opt(bp,cur,TAG_DHCP_RENEW,hp->renew);
			break;
		case TAG_SUBNET_MASK:
			if(f.subnet_mask=hp->flags.subnet_mask)
			rc=insert_ulong_opt(bp,cur,TAG_SUBNET_MASK,hp->subnet_mask.s_addr);
			break;
		case TAG_NB_SCOPE:
			if(f.nb_scope=hp->flags.nb_scope)
			rc=insert_str_opt(bp,cur,TAG_NB_SCOPE,hp->nb_scope->string);
			break;
		case TAG_DOMAIN_NAME:
			if(f.domain_name=hp->flags.domain_name)
				rc=insert_str_opt(bp,cur,TAG_DOMAIN_NAME,hp->domain_name->string);
			else{
				for(j=0;hp->hostname->string[j] && hp->hostname->string[j]!='.';j++);
				rc=insert_str_opt(bp,cur,TAG_DOMAIN_NAME,hp->hostname->string+j+1);
			}
				
			break;
		case TAG_HOST_NAME:
			f.name_switch=hp->flags.name_switch;
			f.send_name=hp->flags.send_name;
			rc=insert_str_opt(bp,cur,TAG_HOST_NAME,hp->hostname->string);
			break;
		case TAG_COOKIE_SERVER:
		    if(f.cookie_server=hp->flags.cookie_server)
			rc=insert_addrs_opt(bp,cur,TAG_COOKIE_SERVER,hp->cookie_server->addr,
							hp->cookie_server->addrcount);
			break;
		case TAG_IMPRESS_SERVER:
		    if(f.impress_server=hp->flags.impress_server)
			rc=insert_addrs_opt(bp,cur,TAG_IMPRESS_SERVER,hp->impress_server->addr,
							hp->impress_server->addrcount);
			break;
		case TAG_LOG_SERVER:
		    if(f.log_server=hp->flags.log_server)
			rc=insert_addrs_opt(bp,cur,TAG_LOG_SERVER,hp->log_server->addr,
							hp->log_server->addrcount);
			break;
		case TAG_LPR_SERVER:
		    if(f.lpr_server=hp->flags.lpr_server)
			rc=insert_addrs_opt(bp,cur,TAG_LPR_SERVER,hp->lpr_server->addr,
							hp->lpr_server->addrcount);
			break;
		case TAG_NAME_SERVER:
		    if(f.name_server=hp->flags.name_server)
			rc=insert_addrs_opt(bp,cur,TAG_NAME_SERVER,hp->name_server->addr,
							hp->name_server->addrcount);
			break;
		case TAG_RLP_SERVER:
		    if(f.rlp_server=hp->flags.rlp_server)
			rc=insert_addrs_opt(bp,cur,TAG_RLP_SERVER,hp->rlp_server->addr,
							hp->rlp_server->addrcount);
			break;
		case TAG_TIME_SERVER:
		    if(f.time_server=hp->flags.time_server)
			rc=insert_addrs_opt(bp,cur,TAG_TIME_SERVER,hp->time_server->addr,
							hp->time_server->addrcount);
			break;
		case TAG_NIS_SERVER:
		    if(f.nis_server=hp->flags.nis_server)
			rc=insert_addrs_opt(bp,cur,TAG_NIS_SERVER,hp->nis_server->addr,
							hp->nis_server->addrcount);
			break;
		case TAG_NTP_SERVER:
		    if(f.ntp_server=hp->flags.ntp_server)
			rc=insert_addrs_opt(bp,cur,TAG_NTP_SERVER,hp->ntp_server->addr,
							hp->ntp_server->addrcount);
			break;
		case TAG_GATEWAY:
		    if(f.gateway=hp->flags.gateway)
			rc=insert_addrs_opt(bp,cur,TAG_GATEWAY,hp->gateway->addr,
							hp->gateway->addrcount);
			break;
		case TAG_DOMAIN_SERVER:
		    if(f.domain_server=hp->flags.domain_server)
			rc=insert_addrs_opt(bp,cur,TAG_DOMAIN_SERVER,hp->domain_server->addr,
							hp->domain_server->addrcount);
			break;
		case TAG_NBNS_SERVER: 
		    if(f.nbns_server=hp->flags.nbns_server)
			rc=insert_addrs_opt(bp,cur,TAG_NBNS_SERVER,hp->nbns_server->addr,
							hp->nbns_server->addrcount);
			break;
		default:if(debug>2)
					report(LOG_INFO,"Unknown option request %d",req_opts[i]);
		}
	}
	return f;

}

void send_dhcpnack(pkt,cliid,msg,max_len)
struct bootp *pkt;
struct shared_bindata *cliid;
int max_len;
{
	unsigned char *opts;
	int32 len;
	unsigned long addr;
	struct opts opt;
	struct bootp *tmp;

	tmp=(struct bootp*)send_buf;
	bcopy(pkt,tmp,pktlen);

	tmp->bp_op=BOOTREPLY;
	tmp->bp_hops=0;
	tmp->bp_secs=0;
	tmp->bp_ciaddr.s_addr=0;
	tmp->bp_siaddr.s_addr=0;
	tmp->bp_yiaddr.s_addr=0;
	tmp->bp_flags|=0x8000;

	opt.cur=0;
	opt.len=max_len;
	insert_byte_opt(tmp,&opt,TAG_DHCP_MSG,DHCPNACK);
	insert_string_opt(tmp,&opt,TAG_DHCP_CLIID,cliid->data,cliid->length);
	insert_str_opt(tmp,&opt,TAG_MESSAGE,msg);
	insert_end_opt(tmp,&opt);
	report(LOG_INFO,"Sending DHCPNACK %s",msg);
	nacks_sent++;
	if(!tmp->bp_giaddr.s_addr)
		addr=0xffffffff;
	else
		addr=tmp->bp_giaddr.s_addr;
	len=300;
	if(opt.cur<VEND_OFFSET)
		len=max_len;
	if(opt.cur>len)
		len=opt.cur;
	send_dhcp_reply(tmp,len,addr);
}

		

void send_dhcp_reply(pkt,send_len,dest)
struct bootp *pkt;
int32 send_len, dest;
{
	struct sockaddr_in dest_addr;
    unsigned char *ha;
    int len, haf;
    struct in_addr dst;

	dest_addr.sin_family = AF_INET;
	dest_addr.sin_port = htons(bootpc_port);
	if(dest)
		dest_addr.sin_addr.s_addr=dest;
	 else if(pkt->bp_ciaddr.s_addr)
		dest_addr.sin_addr.s_addr=pkt->bp_ciaddr.s_addr;
	else if(pkt->bp_giaddr.s_addr){
		dest_addr.sin_addr.s_addr=pkt->bp_giaddr.s_addr;
		dest_addr.sin_port=bootps_port;
	} else {
	dst = pkt->bp_yiaddr;
	ha = pkt->bp_chaddr;
	len = pkt->bp_hlen;
	if (len > MAXHADDRLEN)
	    len = MAXHADDRLEN;
	haf = (int) pkt->bp_htype;
	if (haf == 0)
	    haf = HTYPE_ETHERNET;
	if (debug > 1)
	    report(LOG_INFO, "setarp %s - %s", inet_ntoa(dst), haddrtoa(ha, len));
	setarp(s, &dst, haf, ha, len);
	dest_addr.sin_addr=dst;
	}       
	report(LOG_INFO,"Sending reply %s %d bytes",inet_ntoa(dest_addr.sin_addr),send_len);
	if(sendto(s,(char*)pkt,send_len,0,(struct sockaddr *)&dest_addr,sizeof(dest_addr)) <0)
		 report(LOG_ERR, "sendto: %s", get_network_errmsg());
}



void dhcp_garbage_collect()
{
 unsigned hashcode,iphashcode;
 struct host *hp,*hp2;
 struct order_list *t;
 struct cable *c;


	if(debug>1)
		report(LOG_INFO,"DHCP garbage collect begin");
 hp=(struct host*)hash_FirstEntry(boundhshtbl);
 while(hp && (hp=(struct host*)hash_NextEntry(boundhshtbl))){
	if(hp->expire>now)
		continue;
	ips_expired++;
	if(debug>2)
		report(LOG_INFO,"%s has expired",inet_ntoa(hp->iaddr));
	if(debug>2)
		report(LOG_INFO,"removing binding");
	reuse_addr(hp);
 }
	if(debug>1)
		report(LOG_INFO,"DHCP garbage collect end");
	now=time(0);
	dhcp_alarm=now+gcinterval;
}


struct host *create_lease(hp,cliid,vend,user,secs)
struct host *hp;
struct shared_bindata *cliid,*vend,*user;
unsigned secs;
{
	struct host *hp2;
	unsigned hashcode;
	char override[2000];
	

	if(user){
		override[0]='$';
		override[1]='C';
		override[2]=0;
		for(i=0;i<user->length;i++)
			sprintf(override+strlen(override),"%02X",user->data[i]);
	} else if(vend){
		override[0]='$';
		override[1]='V';
		override[2]=0;
		for(i=0;i<vend->length;i++)
			sprintf(override+strlen(override),"%02X",vend->data[i]);
	} else 
		strcpy(override,"$DHCP");
	hashcode=hash_HashFunction((u_char*)override,strlen(override));
	hp2=(struct host *)hash_Lookup(nmhashtable,hashcode,nmcmp,override);
	if(hp2)
		fill_defaults2(hp,hp2);
	hp2=(struct host*)hash_Lookup(iphashtable,hp->iphashcode,ipcmp,
			&hp->iaddr.s_addr);
	if(!hp2)
		return NULL;
	fill_defaults2(hp,hp2);
	if(!hp->flags.lease){
		hp->flags.lease=TRUE;
		hp->lease=DEFAULT_LEASE;
	}
	if(secs){
		if(hp->flags.max_lease)
			if(secs<hp->flags.max_lease)
				hp->lease=secs;
		else
			if(secs<hp->lease)
				hp->lease=secs;
	}
	if(!hp->flags.renew){
		hp->renew=hp->lease/2;
		hp->flags.renew=TRUE;
	}
	if(!hp->flags.rebind){
		hp->rebind=hp->lease*0.85;
		hp->flags.rebind=TRUE;
	}
		
	hp->flags.complete=TRUE;
	return hp;
}
struct host *create_stub(hp,cl)
struct host *hp;
struct client_info *cl;
{
	struct host *t;
	t=(struct host*)smalloc(sizeof(struct host));
	bzero(t,sizeof(struct host));
	t->linkcount=0;
	t->flags.iaddr=TRUE;
	t->iaddr=hp->iaddr;
	t->hostname=hp->hostname;
	t->hostname->linkcount++;
	t->flags.bootserver=TRUE;
	if(hp->flags.reply_addr)
		t->reply_addr.s_addr=hp->reply_addr.s_addr;
	else
		t->reply_addr.s_addr=0;
	t->flags.reply_addr=TRUE;

	t->cl=cl->cliid;
	t->iphashcode=hp->iphashcode;
	t->cl->linkcount++;

	return t;
}

