/***********************************************************
        Copyright 1996 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "bootp.h"
#include "bootpd.h"
#include "dhcp.h"
#include "dhcpd.h"
#include "hash.h"
#include "perm.h"
#include "syslog.h"
#include "report.h"
#include "readfile.h"

hash_tbl *lastbindhshtbl,*cache2disk;
hashf_tbl *lastbindpermanent;
hashf_tbl *boundpermanent;
int max_cachequeue;
unsigned long flush_interval,flush_alarm;
static int num_in_queue=0;
char *bindingsdir=LASTBINDDIR;



void free_lastbind(a)
hash_datum *a;
{
struct lastbind *p=a;
	if(p==NULL)
		return;
	if(p->linkcount--)
		return;
	del_bindata(p->cliid);
	del_bindata(p->cliidhw);
	del_bindata(p->vend);
	del_bindata(p->user);
	free(p);
}

boolean lastbindcliidcmp(a,b)
hash_datum *a,*b;
{
struct shared_bindata *aa=a;
struct lastbind *ba=b;
	return	BINDATA_EQ(aa,ba->cliid);
}
boolean lastbindcmp(a,b)
hash_datum *a,*b;
{
struct lastbind *aa=a,*ba=b;
	return	BINDATA_EQ(aa->cliid,ba->cliid);
}
boolean bindfilecmp(a,b)
hash_datum *a,*b;
{
struct shared_bindata *aa=a;
unsigned char *ba=b;
	if(ba[0]!=aa->length)
		return FALSE;
	return !bcmp(ba+1,aa->data,ba[0]);
}

void	flush_disk()
{
	int i;
	char buf[MAXELEMENTLEN];	
	unsigned long datalen;
	struct lastbind *p;

	if(debug>1)
		report(LOG_INFO,"Flushing %d bindings to disk",num_in_queue);
	while(p=(struct lastbind*)hash_FirstEntry(cache2disk)){
		datalen=0;
		buf[datalen++]=p->cliid->length;
		bcopy(p->cliid->data,&buf[datalen], p->cliid->length);
		datalen+=p->cliid->length;

		buf[datalen++]=p->cliidhw->length;
		bcopy(p->cliidhw->data,&buf[datalen], p->cliidhw->length);
		datalen+=p->cliidhw->length;

		bcopy(&p->last_ip,&buf[datalen],4);
		datalen+=4;
		bcopy(&p->replys,&buf[datalen],4);
		datalen+=4;
		bcopy(&p->last_touch,&buf[datalen],4);
		datalen+=4;
		bcopy(&p->expire,&buf[datalen],4);
		datalen+=4;
		buf[datalen++]=p->code;

		if(p->vend){
			buf[datalen++]=p->vend->length;
			bcopy(p->vend->data,&buf[datalen], p->vend->length);
			datalen+=p->vend->length;
		} else
			buf[datalen++]=0;

		if(p->user){
			buf[datalen++]=p->user->length;
			bcopy(p->user->data,&buf[datalen], p->user->length);
			datalen+=p->user->length;
		} else
			buf[datalen++]=0;

		hashf_Replace(lastbindpermanent,p->clhashcode,bindfilecmp,
					p->cliid,buf,datalen);
		hash_Delete(cache2disk,p->clhashcode,lastbindcmp,p,free_lastbind);
		num_in_queue--;
	}
	if(debug>1)
		report(LOG_INFO,"Flush complete %d",num_in_queue);
	flush_alarm=time(0)+flush_interval;
}

int release_lastbind(hp)
struct host *hp;
{
	struct lastbind *b,t;
	char queue;
	int rc;
	unsigned hashcode;


	hashcode=hash_HashFunction(hp->cl->data,hp->cl->length);
	t.cliid=hp->cl;
	b=(struct lastbind*)hash_Lookup(lastbindhshtbl,hashcode,lastbindcmp,&t);
	if(!b)
		return;
	b->expire=now;
	/* need a hash_Replace here, but will do a delete insert */
	rc=hash_Delete(cache2disk,b->clhashcode,lastbindcmp,b,free_lastbind);
	if(rc!= 0)
		num_in_queue++;
	b->linkcount++;
	hash_Insert(cache2disk,b->clhashcode,lastbindcmp,b,b);
	if(num_in_queue>=max_cachequeue)
		flush_disk();
}
int update_lastbind(cl,hp)
struct client_info *cl;
struct host *hp;
{
	struct lastbind *b,t;
	char queue;
	int rc;

    queue=0;
	t.cliid=cl->cliid;
	b=(struct lastbind*)hash_Lookup(lastbindhshtbl,cl->clhashcode,lastbindcmp,&t);
	if(!b){
		queue++;
		b=(struct lastbind*)malloc(sizeof(struct lastbind));
		bzero(b,sizeof(struct lastbind));

		b->clhashcode=cl->clhashcode;
		b->cliid=cl->cliid;
		b->cliid->linkcount++;
		b->cliidhw=cl->cliidhw;
		b->cliidhw->linkcount++;
		if(cl->vend){
			b->vend=cl->vend;
			b->vend->linkcount++;
		}
		if(cl->user){
			b->user=cl->user;
			b->user->linkcount++;
		}
		b->last_ip.s_addr=hp->iaddr.s_addr;
		b->expire=hp->expire;
		b->linkcount=1;
		hash_Insert(lastbindhshtbl,b->clhashcode,nullcmp,b,b);
	} else {
		if(!BINDATA_EQ(b->cliidhw,cl->cliidhw)){
			queue++;
			del_bindata(b->cliidhw);
			b->cliidhw=cl->cliidhw;
			b->cliidhw->linkcount++;
		}
		if(cl->vend && (!b->vend || !BINDATA_EQ(b->vend,cl->vend))){
			queue++;
			del_bindata(b->vend);
			b->vend=cl->vend;
			b->vend->linkcount++;
		}
		if(!cl->vend && b->vend){
			del_bindata(b->vend);
			b->vend=NULL;
		}
		if(cl->user && (!b->user || !BINDATA_EQ(b->user,cl->user))){
			queue++;
			del_bindata(b->user);
			b->user=cl->user;
			b->user->linkcount++;
		}
		if(!cl->user && b->user){
			del_bindata(b->user);
			b->user=NULL;
		}
		if(bcmp(&b->last_ip,&hp->iaddr,4)){
			queue++;
			b->last_ip.s_addr=hp->iaddr.s_addr;
		}
		
	}
	b->replys++;
	b->last_touch=now;
	b->expire=hp->expire;
	b->code=hp->code;
	queue++;

	if(queue){	/* test is pointless, but leave for future use */

		/* need a hash_Replace here, but will do a delete insert */
		rc=hash_Delete(cache2disk,b->clhashcode,lastbindcmp,b,free_lastbind);
		if(rc!= 0)
			num_in_queue++;
		b->linkcount++;
		hash_Insert(cache2disk,b->clhashcode,lastbindcmp,b,b);
		if(num_in_queue>=max_cachequeue)
			flush_disk();
	}
		
}
	
